/*----------------------------------------------------------------------
  HGE Benchmark - Roberto Prieto 
  Copyright (C) 2013-2015 MegaStorm Systems
  contact@megastormsystems.com - http://www.megastormsystems.com

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

 ------------------------------------------------------------------------
                                HGE Benchmark Info
    
    - Uses HGE 1.9.0
    - Direct3D 9
    - 32 version

    Changelog:
 
 v1.00 Sept/2015 - First version.
   
 ------------------------------------------------------------------------ */


#include <windows.h>
#include <time.h>
#include "include\hge.h"
#include "include\hgefont.h"
#include "../SDL-Benchmark.h"

// Common vars
typedef struct HGE_Rect
{
    int x,y;
    int w,h;
} HGE_Rect;
static int num_sprites;
static int sprite_w, sprite_h;
static HGE_Rect *positions;
static HGE_Rect *velocities;
static int iStart;
static int iFrames;
SDLBenchParameters SDLBenchParams; // Parameters struct

// HGE vars
HGE *hge;
HTEXTURE **pTextureTable = NULL;
hgeSprite **pSpriteTable = NULL;
int iRealSprites;
//#define DEBUG_RENDERING

int RemoveSprites(int iNumRealSpr)
{
    int i = 0;

    while(i < iNumRealSpr)
    {        
        if(pSpriteTable != NULL)
        {
            if(pSpriteTable[i] != NULL) delete pSpriteTable[i];
            pSpriteTable[i] = NULL;
        }
        if(pTextureTable != NULL)
        {
            if(pTextureTable[i] != NULL) hge->Texture_Free((HTEXTURE&)pTextureTable[i]);
            pTextureTable[i] = NULL;
        }
        ++i;        
    }
    free(pSpriteTable); 
    pSpriteTable = NULL;
    free(pTextureTable);
    pTextureTable = NULL;
    return 0;
}

int LoadSprites(char *file, int iNumRealSpr)
{
    int i = 0;
   
    // Create textures for each sprite using the surface
    int bFlagError = 0;
    pSpriteTable = new hgeSprite *[iNumRealSpr];    
    pTextureTable = new HTEXTURE *[iNumRealSpr];
    while(i < iNumRealSpr)
    {
        pTextureTable[i] = (HTEXTURE*)hge->Texture_Load(SDLBenchParams.szIcon);
        if(!pTextureTable[i]) 
        {
            fprintf(stderr, "Couldn't create texture for sprite %d: %s\n", i, hge->System_GetErrorMessage());
            bFlagError = 1;
        }
        else
        {
            sprite_w = hge->Texture_GetWidth((HTEXTURE&)pTextureTable[i]);
            sprite_h = hge->Texture_GetHeight((HTEXTURE&)pTextureTable[i]);
            // Apply the "colorkey"...
            DWORD *lock = hge->Texture_Lock((HTEXTURE&)pTextureTable[i]);
            if(lock)
            {
                int pixels = sprite_w * sprite_h;
                for(int j = 0; j < pixels; j++)
                {
                    if((lock[j] & 0x00FFFFFF) == 00)
                    {
                        lock[j] &= 0x00FFFFFF; // make this pixel transparent
                    }
                }
                hge->Texture_Unlock((HTEXTURE&)pTextureTable[i]);
            }
            pSpriteTable[i] = new hgeSprite((HTEXTURE&)pTextureTable[i], 0, 0, sprite_w, sprite_h);
            if(!pSpriteTable[i]) 
            {            
                fprintf(stderr, "Couldn't create sprite %d: %s\n", i, hge->System_GetErrorMessage());
                bFlagError = 1;
            }
        }
        ++i;        
    }
    
    if(bFlagError == 1) 
    {
        RemoveSprites(iNumRealSpr);
        return -1;
    }
    return 0;
}

// Timing control
bool FrameFunc()
{
    if(( (float)(clock() - iStart) / 1000) >= SDLBenchParams.dBaseTime) { return true; }
    return false;
}

// Rendering control
bool RenderFunc()
{
    int i, n;
    HGE_Rect *position, *velocity;
    #ifdef DEBUG_RENDERING
     LARGE_INTEGER StartingTime, EndingTime, measure1, measure2;
     QueryPerformanceCounter(&StartingTime);
    #endif

    // Advance one frame
    ++iFrames;

    // Render the scene
    hge->Gfx_BeginScene();
    
    // Clear the screen
    hge->Gfx_Clear(0x00A0A0A0);
    #ifdef DEBUG_RENDERING
     QueryPerformanceCounter(&measure1);
    #endif

    // Move the sprite, bounce at the wall, and draw
    n = 0;
    for (i = 0; i < num_sprites; ++i) 
    {
        position = &positions[i];
        velocity = &velocities[i];
        position->x += velocity->x;
        if((position->x < 0) || (position->x >= (SDLBenchParams.iVideoWidth - sprite_w))) 
        {
            velocity->x = -velocity->x;
            position->x += velocity->x;
        }
        position->y += velocity->y;
        if ((position->y < 0) || (position->y >= (SDLBenchParams.iVideoHeight - sprite_h))) 
        {
            velocity->y = -velocity->y;
            position->y += velocity->y;
        }

        // Render the sprite onto the screen
        if(iRealSprites == 1) pSpriteTable[0]->Render((float)position->x, (float)position->y);
        else pSpriteTable[i]->Render((float)position->x, (float)position->y);
    }
    #ifdef DEBUG_RENDERING
     QueryPerformanceCounter(&measure2);
    #endif

    hge->Gfx_EndScene();
    
    #ifdef DEBUG_RENDERING
     QueryPerformanceCounter(&EndingTime);
     printf("FrameTime: %llu + %llu + %llu = %llu\n", 
        measure1.QuadPart - StartingTime.QuadPart, 
        measure2.QuadPart - measure1.QuadPart, 
        EndingTime.QuadPart - measure2.QuadPart, 
        EndingTime.QuadPart - StartingTime.QuadPart);     
    #endif
    return false;
}

int main(int argc, char *argv[])
{
    int i, iMaxSprites;
    double fFPS_1, fFPS_2;

    // Print information and argument parsing
    parseAndInformation(argc, argv, SDLBENCH_HGE_VERSION" Benchmarks", &SDLBenchParams);
    
    // Max and real sprites
    iMaxSprites = SDLBenchParams.iNumSprites;
    if(iMaxSprites < SDLBenchParams.iNumSprites2) iMaxSprites = SDLBenchParams.iNumSprites2;
    if(SDLBenchParams.iFillRateMode == 0) iRealSprites = iMaxSprites;
    else iRealSprites = 1;

    // Set desired system states and initialize HGE
    hge = hgeCreate(HGE_VERSION);
    hge->System_SetState(HGE_FRAMEFUNC, FrameFunc);
    hge->System_SetState(HGE_RENDERFUNC, RenderFunc);
    hge->System_SetState(HGE_TITLE, SDLBENCH_HGE_VERSION);
    hge->System_SetState(HGE_USESOUND, false);
    hge->System_SetState(HGE_WINDOWED, false);
    hge->System_SetState(HGE_SCREENWIDTH, SDLBenchParams.iVideoWidth);
    hge->System_SetState(HGE_SCREENHEIGHT, SDLBenchParams.iVideoHeight);
    hge->System_SetState(HGE_SCREENBPP, 32);
    //hge->System_SetState(HGE_LOGFILE, "hge.output.log");
    hge->System_SetState(HGE_FPS, HGEFPS_UNLIMITED);

    // Initialize 
    if(!hge->System_Initiate())
    {
        printf("Couldn't initialize HGE system: %s\n", hge->System_GetErrorMessage());
        hge->System_Shutdown();
        hge->Release();
        return -1;
    }
    printf("HGE renderer: direct3d9\n");
    
    // Load the texture and create the sprite
    if(LoadSprites(SDLBenchParams.szIcon, iRealSprites) < 0) 
    {
        fprintf(stderr, "Couldn't load %s: %s\n", SDLBenchParams.szIcon, hge->System_GetErrorMessage());
        hge->System_Shutdown();
        hge->Release();
        return -2;
    }
    
    // Allocate memory for the sprite info and initialize it    
    positions = (HGE_Rect *) malloc(iMaxSprites * sizeof(HGE_Rect));
    velocities = (HGE_Rect *) malloc(iMaxSprites * sizeof(HGE_Rect));
   
    srand((unsigned int)time(NULL));
    for(i = 0; i < iMaxSprites; ++i) 
    {
        positions[i].x = rand() % (SDLBenchParams.iVideoWidth - sprite_w);
        positions[i].y = rand() % (SDLBenchParams.iVideoHeight - sprite_h);
        positions[i].w = sprite_w;
        positions[i].h = sprite_h;
        velocities[i].x = 0;
        velocities[i].y = 0;
        while(!velocities[i].x && !velocities[i].y) 
        {
            velocities[i].x = (rand() % (MAX_SPEED * 2 + 1)) - MAX_SPEED;
            velocities[i].y = (rand() % (MAX_SPEED * 2 + 1)) - MAX_SPEED;
        }
    }

    // ---Main render loop for SDLBenchParms.iNumSprites---    
    num_sprites = SDLBenchParams.iNumSprites;
    iFrames = 0;	
    iStart = clock();	
    hge->System_Start();

    // Print out some timing information
    if(clock() > iStart) 
    {
        fFPS_1 = ((double)iFrames * 1000) / (clock() - iStart);
        printf(" - Moving %d sprites at %2.2f frames per second\n", num_sprites, fFPS_1);
    }	   
    // ------

    // ---Main render loop for SDLBenchParms.iNumSprites2---
    num_sprites = SDLBenchParams.iNumSprites2;
    iFrames = 0;	
    iStart = clock();	
    hge->System_Start();

    // Print out some timing information
    if(clock() > iStart) 
    {
        fFPS_2 = ((double) iFrames * 1000) / (clock() - iStart);
        printf(" - Moving %d sprites at %2.2f frames per second\n\n", num_sprites, fFPS_2);
    }	
    // ------

    // Integrated mode: extended information output
    if(SDLBenchParams.iIntegrated)
    {
        FILE *fp = fopen("hge.table","wt");
        fprintf(fp,"          <tr><td align=center><b><font size=2 color=#000080>%s - Direct3D9", SDLBENCH_HGE_VERSION);
        fprintf(fp,"</font></b></td>");
        fprintf(fp,"<td align=center><font size=2 color=#800000>%2.2f</font></td>",fFPS_1);
        fprintf(fp,"<td align=center><font size=2 color=#800000>%2.2f</font></td></tr>\n",fFPS_2);
        fclose(fp);
        fp = fopen("SDL-Benchmark.graph","at");
        fprintf(fp,"%s (Direct3D9), %2.2f, %2.2f\n", SDLBENCH_HGE_VERSION, fFPS_1, fFPS_2);
        fclose(fp);
    }

    // Close and free memory buffers
    free(positions);
    free(velocities);
    RemoveSprites(iRealSprites);

    // End!
    hge->System_Shutdown();
    hge->Release();      
    return 0;
}

