#include "..\core\hge_plugin.h"

#include <windows.h>

struct CStreamList
{
	HSTREAM				hstream;
	void*				data;
	CStreamList*		next;
};

/*
** HGE Plugin implementation
*/
class HGE_BASS : public HGE_Plugin
{
	public:
		virtual	void		CALL	Release();

		virtual bool		CALL	System_Initiate();

		virtual HEFFECT		CALL	Effect_Load(const char *filename, DWORD size=0);
		virtual void		CALL	Effect_Free(HEFFECT eff);
		virtual HCHANNEL	CALL 	Effect_Play(HEFFECT eff);
		virtual HCHANNEL	CALL 	Effect_PlayEx(HEFFECT eff, int volume=100, int pan=0, float pitch=1.0f, bool loop=false);

		virtual HMUSIC		CALL 	Music_Load(const char *filename, DWORD size=0);
		virtual void		CALL	Music_Free(HMUSIC mus);
		virtual HCHANNEL	CALL 	Music_Play(HMUSIC mus, bool loop, int volume = 100, int order = 0, int row = 0);
		virtual void		CALL	Music_SetAmplification(HMUSIC music, int ampl);
		virtual int			CALL	Music_GetAmplification(HMUSIC music);
		virtual int			CALL	Music_GetLength(HMUSIC music);
		virtual void		CALL	Music_SetPos(HMUSIC music, int order, int row);
		virtual bool		CALL	Music_GetPos(HMUSIC music, int *order, int *row);
		virtual void		CALL	Music_SetInstrVolume(HMUSIC music, int instr, int volume);
		virtual int			CALL	Music_GetInstrVolume(HMUSIC music, int instr);
		virtual void		CALL	Music_SetChannelVolume(HMUSIC music, int channel, int volume);
		virtual int			CALL	Music_GetChannelVolume(HMUSIC music, int channel);

		virtual HSTREAM		CALL	Stream_Load(const char *filename, DWORD size=0);
		virtual void		CALL	Stream_Free(HSTREAM stream);
		virtual HCHANNEL	CALL	Stream_Play(HSTREAM stream, bool loop, int volume = 100);

		virtual void		CALL 	Channel_SetPanning(HCHANNEL chn, int pan);
		virtual void		CALL 	Channel_SetVolume(HCHANNEL chn, int volume);
		virtual void		CALL 	Channel_SetPitch(HCHANNEL chn, float pitch);
		virtual void		CALL 	Channel_Pause(HCHANNEL chn);
		virtual void		CALL 	Channel_Resume(HCHANNEL chn);
		virtual void		CALL 	Channel_Stop(HCHANNEL chn);
		virtual void		CALL 	Channel_PauseAll();
		virtual void		CALL 	Channel_ResumeAll();
		virtual void		CALL 	Channel_StopAll();
		virtual bool		CALL	Channel_IsPlaying(HCHANNEL chn);
		virtual float		CALL	Channel_GetLength(HCHANNEL chn);
		virtual float		CALL	Channel_GetPos(HCHANNEL chn);
		virtual void		CALL	Channel_SetPos(HCHANNEL chn, float fSeconds);
		virtual void		CALL	Channel_SlideTo(HCHANNEL channel, float time, int volume, int pan = -101, float pitch = -1);
		virtual bool		CALL	Channel_IsSliding(HCHANNEL channel);

		// Audio
		HINSTANCE			hBass;
		bool				bSilent;
		CStreamList*		streams;

		bool				_SoundInit();
		void				_SoundDone();
		void				_SetMusVolume(int vol);
		void				_SetStreamVolume(int vol);
		void				_SetFXVolume(int vol);

		static HGE_BASS * _Interface_Get(HGE *lpHGE);

	private:
		HGE_BASS(HGE *lpHGE);

		char				szError[256];
		void				_PostError(char *error);
};

extern HGE_BASS * g_lpHGE;
