/*----------------------------------------------------------------------
  SDL Benchmark - Roberto Prieto 
  Copyright (C) 2013-2015 MegaStorm Systems
  contact@megastormsystems.com - http://www.megastormsystems.com

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

 ------------------------------------------------------------------------
                                SDL Benchmark Info
    
    SDL Benchmark tests the video performance of SDL, SDL2, HGE and SFML libraries.
    Supports two working modes: Real mode and FillRate mode. All the parameters 
    can be configured using arguments.
    Output to HTML with system details, output table and GNUplot graph.

    Changelog:
 
 v1.10 Sept/2015 - Added HGE 1.9.0 and SFML 2.3.1 benchmarks.
                 - Two working modes: Real mode and FillRate mode.
                 - Improves output HTML. Uses gnuplot. Customization using arguments.
 v1.00 Aug/2015 - First version. Replaced the old glSDL-Benchmark and includes SDL and SDL2 tests.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <windows.h>
#include <winerror.h>
#include "SDL-Benchmark.h"

FILE *fLog;

// Prototypes
char *getOSInfo();
char *getProcessorInfo();
char *getMemoryInfo();
char *getVideocardInfo();
int base64_encode_block(const char *, int, char *);

// Main body
int main(int argc, char *argv[])
{
    FILE *fOut;
    SDLBenchParameters SDLBenchParams; // Parameters struct
    char szArgs[512];
    char szCmdLine[512];
    char *szOSInfo, *szCPUInfo, *szMemInfo, *szVideoInfo, *szGraphData = NULL;
    int i = 1;

    // Print information and argument parsing
    parseAndInformation(argc, argv, SDLBENCH_VERSION, &SDLBenchParams);
    strcpy(szArgs, "-integrated ");
    while(i < argc)
    {
        strcat(szArgs, argv[i]);
        strcat(szArgs, " ");
        ++i;
    }
    
    // Warn about the time it will take
    if(MessageBox(NULL,"This test takes about 200 seconds.\nPlease, wait...", SDLBENCH_VERSION, MB_OKCANCEL|MB_ICONQUESTION) == IDCANCEL) 
    {
        printf(" Closing...\n");
        return 0;
    }
    fLog = fopen("SDL-Benchmark.htm","wt");
   
    // Dump information, description, contacts, greetings and sys info on the HTML
    fprintf(fLog, "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html4/strict.dtd\">\n");
    fprintf(fLog, "<html lang=\"en\">\n");
    fprintf(fLog, "  <head>\n");
    fprintf(fLog, "    <meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\">\n");
    fprintf(fLog, "    <title>%s</title>\n", SDLBENCH_VERSION);
    fprintf(fLog, "  </head>\n");
    fprintf(fLog, "  <body>\n");
    fprintf(fLog, "    <div align=center>\n\n");
    fprintf(fLog, "       <table border=1 cellpadding=0 cellspacing=0 style=border-collapse: collapse bordercolor=#0000FF width=700>\n");
    fprintf(fLog, "         <tr><td width=100%%><p align=center style=><b><font color=#000080>%s<br>Copyright (C) 2013-2015 MegaStorm Systems<br>http://www.megastormsystems.com</font></b></td></tr>\n",SDLBENCH_VERSION);
    fprintf(fLog, "         <tr><td width=100%%><blockquote><p align=left><font color=#3333FF size=2>- %s: DirectX Software, DirectX Hardware, glSDL Hardware and WinDib Software.\n", SDLBENCH_SDL_VERSION);
    fprintf(fLog, "           <br>- %s: Direct3D9, Direct3D11, OpenGL and Software renderers (x86 and x64).<br>- %s: Direct3D9 renderer.<br>- %s: OpenGL renderer.</font></p></blockquote></td></tr>\n", SDLBENCH_SDL2_VERSION, SDLBENCH_HGE_VERSION, SDLBENCH_SFML_VERSION);
    fprintf(fLog, "       </table>\n\n");
    fprintf(fLog, "       <table border=0 cellpadding=0 cellspacing=0 width=600>\n");
    fprintf(fLog, "         <tr><td width=100%%><p align=left><font size=2 color=#008080><br>\n");
    szOSInfo = getOSInfo();
    szCPUInfo = getProcessorInfo();
    szMemInfo= getMemoryInfo();
    szVideoInfo = getVideocardInfo();
    fprintf(fLog, "- Screen resolution %dx%dx%dbits<br>\n",SDLBenchParams.iVideoWidth ,SDLBenchParams.iVideoHeight,32);
    if(SDLBenchParams.iFillRateMode == 0) fprintf(fLog, "- One texture for each sprite<br><br>\n");
    else fprintf(fLog, "- Single texture shared for all sprites<br><br>\n");
    fprintf(fLog, "         </font></p></td></tr>\n");
    fprintf(fLog, "       </table>\n\n");
    fprintf(fLog, "       <table border=1 cellpadding=0 cellspacing=0 style=border-collapse: collapse bordercolor=#3333FF width=500>\n");
    fprintf(fLog, "          <tr><td align=center rowspan=2></td><td align=center colspan=2><b><font size=2 color=#000080>Number of sprites</font></b></td></tr>\n");
    fprintf(fLog, "          <tr><td align=center><font size=2 color=#3333FF>%d</font></td><td align=center><font size=2 color=#3333FF>%d</font></td></tr>\n",SDLBenchParams.iNumSprites, SDLBenchParams.iNumSprites2);
    
    // Execute external SDL 1.x 32bits test
    STARTUPINFO si;
    PROCESS_INFORMATION pi;
    ZeroMemory( &si, sizeof(si) );
    si.cb = sizeof(si);
    ZeroMemory( &pi, sizeof(pi) );
    strcpy(szCmdLine,"sdl-bench.bin ");
    strcat(szCmdLine, szArgs);
    if(!CreateProcess( NULL, // No module name (use command line). 
        szCmdLine,        // Command line. 
        NULL,             // Process handle not inheritable. 
        NULL,             // Thread handle not inheritable. 
        FALSE,            // Set handle inheritance to FALSE. 
        0,                // No creation flags. 
        NULL,             // Use parent's environment block. 
        NULL,             // Use parent's starting directory. 
        &si,              // Pointer to STARTUPINFO structure.
        &pi ))     		  // Pointer to PROCESS_INFORMATION structure.
    {
        fprintf(fLog,"Warning: could not execute %s 32bits benchmark<br>", SDLBENCH_SDL_VERSION);
    }
    else
    {
        WaitForSingleObject(pi.hProcess,INFINITE);
        CloseHandle(pi.hProcess);
        CloseHandle(pi.hThread);
    }
    
    // Load external results
    fOut = fopen("sdl.table", "rt");
    if(fOut != NULL)
    {
        fseek(fOut,0L,SEEK_END);
        int iSize = ftell(fOut);
        fseek(fOut,0L,0);
        char *pBuffer = (char*)malloc(iSize + 1);
        memset(pBuffer,0,iSize);
        fread(pBuffer,sizeof(char),iSize,fOut);
        fclose(fOut);
        fwrite(pBuffer,sizeof(char),iSize,fLog);
        free(pBuffer);
        system("del sdl.table");
    }
    else fprintf(fLog,"Warning: could not load %s 32bits benchmark output<br>", SDLBENCH_SDL_VERSION);
    
    // Execute external SDL 2.x 32bits test
    _putenv("SDL_VIDEODRIVER=");
    ZeroMemory( &si, sizeof(si) );
    si.cb = sizeof(si);
    ZeroMemory( &pi, sizeof(pi) );
    strcpy(szCmdLine,"sdl2-bench.bin ");
    strcat(szCmdLine, szArgs);
    if(!CreateProcess( NULL, // No module name (use command line). 
        szCmdLine,        // Command line. 
        NULL,             // Process handle not inheritable. 
        NULL,             // Thread handle not inheritable. 
        FALSE,            // Set handle inheritance to FALSE. 
        0,                // No creation flags. 
        NULL,             // Use parent's environment block. 
        NULL,             // Use parent's starting directory. 
        &si,              // Pointer to STARTUPINFO structure.
        &pi ))     		  // Pointer to PROCESS_INFORMATION structure.
    {
        fprintf(fLog,"Warning: could not execute %s 32bits benchmark<br>", SDLBENCH_SDL2_VERSION);
    }
    else
    {
        WaitForSingleObject(pi.hProcess,INFINITE);
        CloseHandle(pi.hProcess);
        CloseHandle(pi.hThread);
    }
    
    // Load external results
    fOut = fopen("sdl2.table", "rt");
    if(fOut != NULL)
    {
        fseek(fOut,0L,SEEK_END);
        int iSize = ftell(fOut);
        fseek(fOut,0L,0);
        char *pBuffer = (char*)malloc(iSize + 1);
        memset(pBuffer,0,iSize);
        fread(pBuffer,sizeof(char),iSize,fOut);
        fclose(fOut);
        fwrite(pBuffer,sizeof(char),iSize,fLog);
        free(pBuffer);
        system("del sdl2.table");
    }
    else fprintf(fLog,"Warning: could not load %s 32bits benchmark output<br>", SDLBENCH_SDL2_VERSION);

    // Execute external SDL 2.x 64bits test
    SYSTEM_INFO sinfo;
    memset(&si, 0, sizeof(sinfo));
    GetNativeSystemInfo(&sinfo);
    if(sinfo.wProcessorArchitecture == PROCESSOR_ARCHITECTURE_AMD64)
    {
        _putenv("SDL_VIDEODRIVER=");
        ZeroMemory( &si, sizeof(si) );
        si.cb = sizeof(si);
        ZeroMemory( &pi, sizeof(pi) );
        strcpy(szCmdLine,"sdl2-bench64.bin ");
        strcat(szCmdLine, szArgs);
        if(!CreateProcess( NULL, // No module name (use command line). 
            szCmdLine,        // Command line. 
            NULL,             // Process handle not inheritable. 
            NULL,             // Thread handle not inheritable. 
            FALSE,            // Set handle inheritance to FALSE. 
            0,                // No creation flags. 
            NULL,             // Use parent's environment block. 
            NULL,             // Use parent's starting directory. 
            &si,              // Pointer to STARTUPINFO structure.
            &pi ))     		  // Pointer to PROCESS_INFORMATION structure.
        {
            fprintf(fLog,"Warning: could not execute %s 64bits benchmark<br>", SDLBENCH_SDL2_VERSION);
        }
        else
        {
            WaitForSingleObject(pi.hProcess,INFINITE);
            CloseHandle(pi.hProcess);
            CloseHandle(pi.hThread);
        }
    
        // Load external results
        fOut = fopen("sdl2-64.table", "rt");
        if(fOut != NULL)
        {
            fseek(fOut,0L,SEEK_END);
            int iSize = ftell(fOut);
            fseek(fOut,0L,0);
            char *pBuffer = (char*)malloc(iSize + 1);
            memset(pBuffer,0,iSize);
            fread(pBuffer,sizeof(char),iSize,fOut);
            fclose(fOut);
            fwrite(pBuffer,sizeof(char),iSize,fLog);
            free(pBuffer);
            system("del sdl2-64.table");
        }
        else fprintf(fLog,"Warning: could not load %s 64bits benchmark output<br>", SDLBENCH_SDL2_VERSION);
    }
    
    // Execute external HGE 1.x 32bits test
    ZeroMemory( &si, sizeof(si) );
    si.cb = sizeof(si);
    ZeroMemory( &pi, sizeof(pi) );
    strcpy(szCmdLine,"hge-bench.bin ");
    strcat(szCmdLine, szArgs);
    if(!CreateProcess( NULL, // No module name (use command line). 
        szCmdLine,        // Command line. 
        NULL,             // Process handle not inheritable. 
        NULL,             // Thread handle not inheritable. 
        FALSE,            // Set handle inheritance to FALSE. 
        0,                // No creation flags. 
        NULL,             // Use parent's environment block. 
        NULL,             // Use parent's starting directory. 
        &si,              // Pointer to STARTUPINFO structure.
        &pi ))     		  // Pointer to PROCESS_INFORMATION structure.
    {
        fprintf(fLog,"Warning: could not execute %s 32bits benchmark<br>", SDLBENCH_HGE_VERSION);
    }
    else
    {
        WaitForSingleObject(pi.hProcess,INFINITE);
        CloseHandle(pi.hProcess);
        CloseHandle(pi.hThread);
    }
    
    // Load external results
    fOut = fopen("hge.table", "rt");
    if(fOut != NULL)
    {
        fseek(fOut,0L,SEEK_END);
        int iSize = ftell(fOut);
        fseek(fOut,0L,0);
        char *pBuffer = (char*)malloc(iSize + 1);
        memset(pBuffer,0,iSize);
        fread(pBuffer,sizeof(char),iSize,fOut);
        fclose(fOut);
        fwrite(pBuffer,sizeof(char),iSize,fLog);
        free(pBuffer);
        system("del hge.table");
    }
    else fprintf(fLog,"Warning: could not load %s 32bits benchmark output<br>", SDLBENCH_HGE_VERSION);
    
    // Execute external SFML 2.x 32bits test
    ZeroMemory( &si, sizeof(si) );
    si.cb = sizeof(si);
    ZeroMemory( &pi, sizeof(pi) );
    strcpy(szCmdLine,"sfml-bench.bin ");
    strcat(szCmdLine, szArgs);
    if(!CreateProcess( NULL, // No module name (use command line). 
        szCmdLine,        // Command line. 
        NULL,             // Process handle not inheritable. 
        NULL,             // Thread handle not inheritable. 
        FALSE,            // Set handle inheritance to FALSE. 
        0,                // No creation flags. 
        NULL,             // Use parent's environment block. 
        NULL,             // Use parent's starting directory. 
        &si,              // Pointer to STARTUPINFO structure.
        &pi ))     		  // Pointer to PROCESS_INFORMATION structure.
    {
        fprintf(fLog,"Warning: could not execute %s 32bits benchmark<br>", SDLBENCH_SFML_VERSION);
    }
    else
    {
        WaitForSingleObject(pi.hProcess,INFINITE);
        CloseHandle(pi.hProcess);
        CloseHandle(pi.hThread);
    }
    
    // Load external results
    fOut = fopen("sfml.table", "rt");
    if(fOut != NULL)
    {
        fseek(fOut,0L,SEEK_END);
        int iSize = ftell(fOut);
        fseek(fOut,0L,0);
        char *pBuffer = (char*)malloc(iSize + 1);
        memset(pBuffer,0,iSize);
        fread(pBuffer,sizeof(char),iSize,fOut);
        fclose(fOut);
        fwrite(pBuffer,sizeof(char),iSize,fLog);
        free(pBuffer);
        system("del sfml.table");
    }
    else fprintf(fLog,"Warning: could not load %s 32bits benchmark output<br>", SDLBENCH_SFML_VERSION);
    
    // Create the gnuplot script    
    fOut = fopen("SDL-Benchmark.gpi", "wt");
    fprintf(fOut, "# GNUPlot settings created for %s\nset timestamp\nset terminal png truecolor size 800,500 enhanced font \"arial,9\"\n", SDLBENCH_VERSION);
    fprintf(fOut, "set style fill solid\nset style data histograms\nset grid y\n");    
    fprintf(fOut, "set output \"SDL-Benchmark.png\"\nset ylabel \"Frames per second\"\nset boxwidth 0.8\nset xtics rotate by -45\nset xtics font \"arial,8\"\nset xtics offset -1.5\n");
    fprintf(fOut, "set title \"%s - %s - %s\\n{/=8 %s} \\n{/=8 %s - %dx%d - ", SDLBENCH_VERSION, szOSInfo, szCPUInfo, szVideoInfo, szMemInfo,SDLBenchParams.iVideoWidth ,SDLBenchParams.iVideoHeight);
    if(SDLBenchParams.iFillRateMode == 0) fprintf(fOut,"One texture for each sprite}\"\n");
    else fprintf(fOut, "Single texture shared for all sprites}\"\n");
    fprintf(fOut, "set datafile separator \",\"\n");
    fprintf(fOut, "plot 'SDL-Benchmark.graph' using 2:xtic(1) lt rgb \"#406090\" title \"%d Sprites\", '' using 3:xtic(1) lt rgb \"#4060F0\" title \"%d Sprites\"\nexit", SDLBenchParams.iNumSprites, SDLBenchParams.iNumSprites2);
    fclose(fOut);

    // Call gnuplot and generate the png
    if(system("gnuplot SDL-Benchmark.gpi") == 0)
    {
        // We have the PNG so work on it
        fOut = fopen("SDL-Benchmark.png","rb");
        if(fOut != NULL)
        {
            // Read the PNG and encode it using base64, szGraph will have the data and i the size
            fseek(fOut,0L,SEEK_END);
            i = ftell(fOut);
            fseek(fOut,0L,0);
            char *pSrc = (char*)malloc(i);
            szGraphData = (char*)malloc(i * 2);
            fread(pSrc, sizeof(char), i, fOut);        
            i = base64_encode_block(pSrc, i, szGraphData);
            free(pSrc);
            fclose(fOut);
        }
    }

    // Finalize the HTML output, close handler and remove temporal files	
    fprintf(fLog, "       </table>\n\n");
    fprintf(fLog, "       <table border=0 cellpadding=0 cellspacing=0 width=800>\n");
    fprintf(fLog, "         <tr><td width=100%%>\n");
    if(szGraphData != NULL) // Dump base64 encoded PNG
    {        
        fprintf(fLog, "           <img src=\"data:SDL-Benchmark/png;base64,");
        fwrite(szGraphData, i, 1, fLog);
        fprintf(fLog, "\" />\n");
        free(szGraphData);
        szGraphData = NULL;
    }
    else // Dump an information message of gnuplot not found
    {
        fprintf(fLog, "<br><font size=2 color=#940909><p align=center>Can not produce a graph image with the results: gnuplot executable not found on path.<br>Please, visit <a href=\"http://www.gnuplot.info/\"> gnuplot home page</a> for downloading and installing it.</p></font>");
    }
    fprintf(fLog, "         </td></tr>\n       </table>\n\n    </div>\n  </body>\n</html>");
    fclose(fLog);    
    system("del SDL-Benchmark.gpi");
    system("del SDL-Benchmark.graph");
    system("del SDL-Benchmark.png");
    return 0;
}

// --- Support functions to get system information ---

// Remove whitespace from a char
void trim(char * s) 
{
    char * p = s;
    int l = strlen(p);

    while(isspace(p[l - 1])) p[--l] = 0;
    while(* p && isspace(* p)) ++p, --l;

    memmove(s, p, l + 1);
}


// Get OS info
char *getOSInfo()
{
    LONG result;
    HKEY hKey;
    static TCHAR OSData[128];
    DWORD dataSize;

    // Get the OS name
    strcpy(OSData,"Unknown OS");
    result = ::RegOpenKeyEx (HKEY_LOCAL_MACHINE,"Software\\Microsoft\\Windows NT\\CurrentVersion", 0, KEY_QUERY_VALUE, &hKey);

    // Check if the function has succeeded.
    if(result == ERROR_SUCCESS) 
    {
        dataSize = sizeof(OSData);
        result = ::RegQueryValueEx (hKey, "ProductName", NULL, NULL,(LPBYTE)OSData, &dataSize);
        trim(OSData);
    }

    // Make sure to close the reg key
    RegCloseKey(hKey);
    // Print info
    fprintf(fLog,"- %s<br>\n", OSData);
    return OSData;
}

// Get cpu name
char *getProcessorInfo()
{
    LONG result;
    HKEY hKey;
    static TCHAR vendorData[128];
    DWORD dataSize;

    // Get the processor name
    strcpy(vendorData,"Unknown CPU Processor");
    result = ::RegOpenKeyEx (HKEY_LOCAL_MACHINE,"Hardware\\Description\\System\\CentralProcessor\\0", 0, KEY_QUERY_VALUE, &hKey);

    // Check if the function has succeeded.
    if(result == ERROR_SUCCESS) 
    {
        dataSize = sizeof(vendorData);
        result = ::RegQueryValueEx (hKey, "ProcessorNameString", NULL, NULL,(LPBYTE)vendorData, &dataSize);
        trim(vendorData);
    }

    // Make sure to close the reg key
    RegCloseKey(hKey);
    // Print info
    fprintf(fLog,"- %s<br>\n", vendorData);
    return vendorData;
}

// Get system memory
char *getMemoryInfo()
{
    int iTmp;
    char sTmp[32];
    static char sMem[64];
    MEMORYSTATUSEX memoryStatus;    

    memoryStatus.dwLength = sizeof(memoryStatus);
    GlobalMemoryStatusEx(&memoryStatus);
    strcpy(sMem,"Memory ");
    iTmp = memoryStatus.ullAvailPhys / 1024 / 1024; // MBytes
    sprintf(sTmp,"%d",iTmp);
    strcat(sMem,sTmp);
    strcat(sMem,"MB / ");

    iTmp = memoryStatus.ullTotalPhys / 1024 / 1024; // MBytes
    sprintf(sTmp,"%d",iTmp);
    strcat(sMem,sTmp);
    strcat(sMem,"MB");

    // Print info
    fprintf(fLog,"- %s<br>\n",sMem);
    return sMem;
}

// Use SDL 1.2 for getting video card information
#include "SDL12.experimental\include\SDL.h"
#include "SDL12.experimental\include\SDL_opengl.h"
char *getVideocardInfo()
{
    static char sVideo[256];
    
    // Initialize SDL
    if(SDL_Init(SDL_INIT_VIDEO) < 0) return false;

    if(SDL_SetVideoMode(640, 480, 32,SDL_OPENGL|SDL_NOFRAME) == NULL) strcpy(sVideo,"Unknown Video card");
    else
    {
        strcpy(sVideo, " ");
        strcat(sVideo, (const char*)glGetString(GL_RENDERER));
        strcat(sVideo, " ");
        strcat(sVideo, (const char*)glGetString(GL_VERSION));		
    }

    // print info
    SDL_Quit();
    fprintf(fLog,"- %s<br>\n",sVideo);
    return sVideo;
}

// ---Encode base64 function---
const int CHARS_PER_LINE = 72;

typedef enum
{
    step_A, step_B, step_C
} base64_encodestep;

typedef struct
{
    base64_encodestep step;
    char result;
    int stepcount;
} base64_encodestate;

char base64_encode_value(char value_in)
{
    static const char* encoding = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
    if (value_in > 63) return '=';
    return encoding[(int)value_in];
}

void base64_init_encodestate(base64_encodestate* state_in)
{
    state_in->step = step_A;
    state_in->result = 0;
    state_in->stepcount = 0;
}

int base64_encode_block(const char* plaintext_in, int length_in, char* code_out)
{
    const char* plainchar = plaintext_in;
    const char* const plaintextend = plaintext_in + length_in;
    char* codechar = code_out;
    char result;
    char fragment;
    base64_encodestate _state;
    base64_init_encodestate(&_state);

    result = _state.result;
    
    switch (_state.step)
    {
        while (1)
        {
    case step_A:
            if (plainchar == plaintextend)
            {
                _state.result = result;
                _state.step = step_A;
                return codechar - code_out;
            }
            fragment = *plainchar++;
            result = (fragment & 0x0fc) >> 2;
            *codechar++ = base64_encode_value(result);
            result = (fragment & 0x003) << 4;
    case step_B:
            if (plainchar == plaintextend)
            {
                _state.result = result;
                _state.step = step_B;
                return codechar - code_out;
            }
            fragment = *plainchar++;
            result |= (fragment & 0x0f0) >> 4;
            *codechar++ = base64_encode_value(result);
            result = (fragment & 0x00f) << 2;
    case step_C:
            if (plainchar == plaintextend)
            {
                _state.result = result;
                _state.step = step_C;
                return codechar - code_out;
            }
            fragment = *plainchar++;
            result |= (fragment & 0x0c0) >> 6;
            *codechar++ = base64_encode_value(result);
            result  = (fragment & 0x03f) >> 0;
            *codechar++ = base64_encode_value(result);
            
            ++(_state.stepcount);
            if (_state.stepcount == CHARS_PER_LINE/4)
            {
                *codechar++ = '\n';
                _state.stepcount = 0;
            }
        }
    }
    /* control should not reach here */
    return codechar - code_out;
}